import logging
from typing import List

from tickthon import Task


class Bulloh:

    @staticmethod
    def process_weight(date: str, weight_logs: List[Task]) -> float:
        """Processes weight data from Ticktick and returns the weight for the given date"""
        logging.info(f"Processing weight data for date: {date}")

        raw_weight = next(filter(lambda log: date in log.created_date, weight_logs), None)
        if raw_weight:
            return float(raw_weight.title.split()[-1])
        return 0.0



import os
from datetime import date

import pytest

from bulloh.rescuetime_client import RescuetimeClient

current_date = date.today().strftime("%Y-%m-%d")


@pytest.fixture
def rescuetime_client():
    return RescuetimeClient(os.getenv("RT_API_KEY"))


def test_get_productivity_time(rescuetime_client):
    productivity_time = rescuetime_client.get_productive_time(current_date)

    assert isinstance(productivity_time, float)
    assert productivity_time >= 0


def test_get_leisure_time(rescuetime_client):
    leisure_time = rescuetime_client.get_leisure_time(current_date)

    assert isinstance(leisure_time, float)
    assert leisure_time >= 0


def test_wrong_date(rescuetime_client):
    wrong_date = "error_date"

    with pytest.raises(KeyError):
        rescuetime_client.get_productive_time(wrong_date)


def test_wrong_credentials():
    with pytest.raises(KeyError):
        RescuetimeClient("").get_productive_time(current_date)



import math

import pytest
from tickthon import Task

from bulloh import Bulloh


@pytest.fixture
def bulloh():
    return Bulloh()


def test_process_weight(bulloh):
    expected_weight = 99.9
    date = "2099-09-09"
    weight_logs = [Task(title="test weright wrong", ticktick_id="", ticktick_etag="", created_date="2099-01-01"),
                   Task(title="weight measurement: 99.9", ticktick_id="", ticktick_etag="", created_date="2099-09-09"),
                   Task(title="test weright wrong", ticktick_id="", ticktick_etag="", created_date="2099-01-01"),]

    weight = bulloh.process_weight(date, weight_logs)

    assert isinstance(weight, float)
    assert math.isclose(weight, expected_weight)


def test_process_weight_without_logs(bulloh):
    expected_weight = 0.0
    date = "2099-09-09"
    weight_logs = [Task(title="weight measurement: 99.9", ticktick_id="", ticktick_etag="", created_date="2099-01-01"),
                   Task(title="weight measurement: 99.9", ticktick_id="", ticktick_etag="", created_date="2099-01-02"),
                   Task(title="weight measurement: 99.9", ticktick_id="", ticktick_etag="", created_date="2099-01-03")]

    weight = bulloh.process_weight(date, weight_logs)

    assert isinstance(weight, float)
    assert math.isclose(weight, expected_weight)



from datetime import datetime, timedelta
import json
import logging
from attr import asdict, define
import boto3
import botocore
import cattrs

from nothion import PersonalStats


@define
class HealthConnectData:
    token: str | None = None
    refresh_token: str | None = None


@define
class TicktickData:
    token: str | None = None
    cookies: dict[str, str] | None = None


@define
class StorageFile:
    health_connect_data: HealthConnectData
    ticktick_data: TicktickData
    stats_data: list[PersonalStats]


class S3Client:

    BASE_FILE_NAME = "bulloh_stats"

    def __init__(self, bucket_name: str):
        self.bucket_name = bucket_name
        self.s3 = boto3.client("s3")

    def upload_file(self, file_content: StorageFile, date: str):
        self.s3.put_object(Bucket=self.bucket_name,
                           Key=f"{self.BASE_FILE_NAME}_{date}.json",
                           Body=json.dumps(asdict(file_content)))

    def download_file(self, date: str) -> StorageFile:
        """Downloads the file from S3 and converts it to a StorageFile object.

        Args:
            date (str): The date to download the file for in the format YYYY-MM-DD.

        Returns:
            StorageFile: The StorageFile object.
        """
        converter = cattrs.Converter()
        converter.register_structure_hook(HealthConnectData, lambda d, _: HealthConnectData(**d))
        converter.register_structure_hook(TicktickData, lambda d, _: TicktickData(**d))
        converter.register_structure_hook(PersonalStats, lambda d, _: PersonalStats(**d))

        file_data = self._try_get_file(date, converter)
        if file_data is not None:
            return file_data

        yesterday = (datetime.strptime(date, "%Y-%m-%d") - timedelta(days=1)).strftime("%Y-%m-%d")
        file_data = self._try_get_file(yesterday, converter)
        if file_data is not None:
            file_data.stats_data = []
            return file_data

        return StorageFile(
            health_connect_data=HealthConnectData(None, None),
            ticktick_data=TicktickData(None, None),
            stats_data=[]
        )

    def _try_get_file(self, date: str, converter) -> StorageFile | None:
        """Helper method to try retrieving a file for a specific date.

        Returns:
            StorageFile or None if file not found.
        """
        try:
            response = self.s3.get_object(
                Bucket=self.bucket_name,
                Key=f"{self.BASE_FILE_NAME}_{date}.json"
            )
            data = json.loads(response["Body"].read().decode("utf-8"))
            return converter.structure(data, StorageFile)
        except self.s3.exceptions.NoSuchKey:
            logging.info(f"No file found for date: {date}")
            return None
        except botocore.exceptions.ClientError as e:
            logging.error(f"Error downloading file: {e}")
            raise e



import logging
from typing import List

import requests


class RescuetimeClient:

    def __init__(self, api_key: str):
        self.api_key = api_key
        self.BASE_URL = f"https://www.rescuetime.com/anapi/data?key={self.api_key}"

    def _get_raw_recorded_hours(self, date: str) -> List[list]:
        """Get the raw recorded hours for a given date.

        Args:
            date: The date to get the recorded hours for in the format YYYY-MM-DD.

        Returns:
            A list of the recorded hours for the given date, each element is a list with the following format:
            [Date, Time Spent (seconds), User ID, Productivity Level (from -2 to 2)]
        """
        recorded_hours_url = f"{self.BASE_URL}&perspective=interval&restrict_kind=productivity&interval=hour&"\
                             f"restrict_begin={date}&restrict_end={date}&format=json"

        return requests.get(recorded_hours_url).json()["rows"]

    def get_productive_time(self, date: str) -> float:
        """Get the productive time for a given date.

        Args:
            date: The date to get the productive time in the format YYYY-MM-DD.
        """
        logging.info(f"Getting productive time for {date}")
        raw_records = self._get_raw_recorded_hours(date)
        return round(sum([record[1] for record in raw_records if record[-1] > 0]) / 3600, 2)

    def get_leisure_time(self, date: str) -> float:
        """Get the leisure time for a given date.

        Args:
            date: The date to get the productive time in the format YYYY-MM-DD.
        """
        logging.info(f"Getting leisure time for {date}")
        raw_records = self._get_raw_recorded_hours(date)
        return round(sum([record[1] for record in raw_records if record[-1] < 0]) / 3600, 2)



from datetime import datetime, timedelta
import json

import requests


class HealthConnectClient:

    BASE_URL = "https://api.hcgateway.shuchir.dev/api/v2"

    def __init__(self, username: str, password: str, api_token: str | None = None, refresh_token: str | None = None):
        self.token, self.refresh_token = self.validate_token(username, password, api_token, refresh_token)
        self.headers = {"Content-Type": "application/json",
                        "Authorization": f"Bearer {self.token}"}

    def _login(self, username: str, password: str):
        """Login to HealthConnect and return the token and refresh token."""
        payload = json.dumps({"username": username, "password": password})
        headers = {"Content-Type": "application/json"}

        response = requests.post(self.BASE_URL + "/login", headers=headers, data=payload)
        response.raise_for_status()

        response_data = response.json()
        return response_data["token"], response_data["refresh"]

    def validate_token(self,
                       username: str,
                       password: str,
                       api_token: str | None,
                       refresh_token: str | None) -> tuple[str, str]:
        """Validate the token and refresh token. If the token is invalid, try to refresh it.
           If the refresh token is invalid, login again.

        Args:
            api_token: The api token to validate.
            refresh_token: The refresh token to validate.
            username: The username to login with.
            password: The password to login with.

        Returns:
            A tuple with the token and refresh token.
        """
        headers = {"Content-Type": "application/json",
                   "Authorization": f"Bearer {api_token}"}
        payload = json.dumps({"queries": {}})

        current_token_response = requests.post(self.BASE_URL + "/fetch/steps", headers=headers, data=payload)
        if current_token_response.ok and api_token and refresh_token:
            return api_token, refresh_token

        if refresh_token:
            headers = {"Content-Type": "application/json"}
            payload = json.dumps({"refresh": refresh_token})
            response = requests.post(self.BASE_URL + "/refresh", headers=headers, data=payload)

            if response.ok:
                return response.json()["token"], response.json()["refresh"]

        return self._login(username, password)

    @staticmethod
    def _get_request_query(date: str = "", app_id: str = "") -> dict:
        query: dict = {"queries": {}}

        if date:
            end_filter_date = datetime.strptime(date, "%Y-%m-%d") + timedelta(days=1)
            query["queries"]["start"] = {"$gte": f"{date}T05:00:00Z",
                                         "$lt": f"{end_filter_date}T05:00:00Z"}

        if app_id:
            query["queries"]["app"] = app_id

        return query

    def get_steps(self, date: str, app_id: str) -> int:
        """Get the number of steps for a given date.

        Args:
            date: The date to get the steps for in the format YYYY-MM-DD.
            app_id: The app id to get the steps for.

        Returns:
            The number of steps for the given date.
        """
        payload = self._get_request_query(app_id=app_id)
        end_date = datetime.strptime(date, "%Y-%m-%d") + timedelta(days=1)
        payload["queries"]["start"] = {"$gte": f"{date}T09:00:00Z",
                                       "$lt": f"{end_date}T04:00:00Z"}

        response = requests.post(self.BASE_URL + "/fetch/steps", headers=self.headers, data=json.dumps(payload))
        response.raise_for_status()

        total_steps = 0
        for step_object in response.json():
            total_steps += step_object["data"]["count"]

        return total_steps

    def get_sleep(self, date: str, app_id: str, sleep_stages: list[int]) -> tuple[float | None, dict[int, float]]:
        """Get the number of sleep for a given date.

        Args:
            date: The date to get the sleep for in the format YYYY-MM-DD.
            app_id: The app id to get the sleep for.
            sleep_stages: The sleep stages to get the sleep for. Official sleep stages are:
                            - 0: UNKNOWN
                            - 1: AWAKE
                            - 2: SLEEPING
                            - 3: OUT_OF_BED
                            - 4: LIGHT
                            - 5: DEEP
                            - 6: REM
                            - 7: AWAKE_IN_BED

        Returns:
            A tuple with the fall asleep time and a dictionary with the sleep stages as keys
            and the number of sleep hours for the given date as values.
        """
        payload = self._get_request_query(app_id=app_id)
        payload["queries"]["start"] = {"$gte": f"{date}T01:00:00Z",
                                       "$lt": f"{date}T15:00:00Z"}

        response = requests.post(self.BASE_URL + "/fetch/sleepSession", headers=self.headers, data=json.dumps(payload))
        response.raise_for_status()

        response_data = response.json()
        all_sleep_stages = []
        for sleep_chunk in response_data:
            all_sleep_stages.extend(sleep_chunk["data"]["stages"])

        total_sleep = {v: 0.0 for v in sleep_stages}
        for stage in all_sleep_stages:
            stage_type = stage["stage"]
            if stage_type not in sleep_stages:
                continue

            start_time = datetime.fromisoformat(stage["startTime"].replace('Z', '+00:00'))
            end_time = datetime.fromisoformat(stage["endTime"].replace('Z', '+00:00'))

            sleep_duration = (end_time - start_time).total_seconds()

            total_sleep[stage_type] += sleep_duration

        total_sleep = {k: round(v / 3600, 2) for k, v in total_sleep.items()}

        if not all_sleep_stages:
            return None, total_sleep

        fall_asleep_raw_time = datetime.fromisoformat(all_sleep_stages[0]["startTime"].replace('Z', '+00:00'))
        fall_asleep_time = (fall_asleep_raw_time - timedelta(hours=5)).strftime("%I.%M")

        return float(fall_asleep_time), total_sleep
